/*=====================================================================
 *
 * VNCCam, a VNC camera frameserver
 * (c) 2001 Michael Rothwell, rothwell@utenzi.com
 * 
 * Licensed under the GPL, see the file "COPYING"
 * for details
 *
 *====================================================================*/

#include <glib.h>
#include <gdk/gdk.h>
#include <ghttp.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>
#include <fcntl.h>
#include <gnome-xml/parser.h>
#include <gnome-xml/tree.h>
#include <gdk-pixbuf/gdk-pixbuf.h>
#include <gdk-pixbuf/gdk-pixbuf-loader.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/tcp.h>
#include <netinet/in.h>
#include <sys/stat.h>
#include <Hermes/Hermes.h>
#include "rfb.h"
#include "rfbproto.h"
#include <signal.h>
#include "textlib.h"
#include "jpegutil.h"


/* BEGIN, defines  ----- ----- ----- ----- ----- ----- ----- ----- ----- ----- -----*/
#define MSG_IMAGE 1
#define	MSG_COPY 2
#define MAX_UPDATE_QUEUE 8		/* more than this are combined into one update */
#define SMALL_TILE 32				/* it seems that the Unix client doesn't like smaller tiles */
#define	SEND_CHUNK 4096				/* amount of data to send to a client at one time */
/* END, defines  ----- ----- ----- ----- ----- ----- ----- ----- ----- ----- -----*/

/* BEGIN, data types  ----- ----- ----- ----- ----- ----- ----- ----- ----- ----- -----*/
typedef struct
{
	void 			*next;
	void 			*prev;
	int				x,y,w,h;
	int				t;
} updates_t;

typedef struct 
{
	int 						x;
	char 						*url;
	char 						*name;
	char						*jpegout;
	int 						y;
	int 						w;
	int 						h;
	GdkPixbuf				*pixbuf;
	HermesFormat 		hformat;
	HermesHandle		hhandle;
	ghttp_request 	*request;
	int							http_p_count;
	guint 					timer_id;
	gboolean				rotate;
} camera_t;

typedef struct 
{
	rfbSetPixelFormatMsg 	rfbformat;
	rfbClientInitMsg 			cim;
	HermesFormat 					hformat;
	HermesHandle					hhandle;
	int										socket;
	int										last_time;
	guint									timer_id;
	gboolean							sending;
	int										x,y,w,h;
	updates_t							*updates;
	char									*upd_buffer;
	int										upd_buffer_len;
	int										upd_buffer_pos;
	int										num_updates;
	int										client_intialized;	/* when we get the first update request, this is set */
} client_t;

typedef struct
{
	void *next;
	void *prev;
	camera_t *this;
} cameras_t;

typedef struct
{
	void *next;
	void *prev;
	client_t *this;
} clients_t;

/* END, data types ----- ----- ----- ----- ----- ----- ----- ----- ----- ----- -----*/



/* START, prototypes  ----- ----- ----- ----- ----- ----- ----- ----- ----- -----*/
gboolean check_req(gpointer data);
void dosendchunk(int x, int y, int w, int h, client_t *client);
void dosendcopyrect(int x, int y, int w, int h, client_t *client);
gboolean do_update(gpointer data);
gboolean do_interval_update(gpointer data);
void run_g_q();
gboolean is_client_connected(client_t *client);
void drawstring(typeutil *t, char *text, unsigned char *bitmap, int w, int h, int bpp, int x, int y);
void dosendscreenblank(client_t *client);
int __WriteExact(client_t *client);
int __WriteExactD(client_t *client,int line)
{
	printf("WRITEEXACT, %d\n",line);
	__WriteExact(client);
};

#ifdef DEBUG
#	define WriteExact(x) __WriteExactD(client,__LINE__)
#else
#	define WriteExact(x) __WriteExact(client)
#endif
/* END, prototypes  ----- ----- ----- ----- ----- ----- ----- ----- ----- -----*/



/* START globals  ----- ----- ----- ----- ----- ----- ----- ----- ----- ----- ----- -----*/
xmlDocPtr					prefs;
GdkPixbuf					*main_canvas;
int								canvas_w;
int								canvas_h;
cameras_t					*camlist=NULL;
clients_t					*clientlist=NULL;
rfbServerInitMsg	sim;
char							*servername;
HermesHandle			conv;
HermesFormat			fmt_source;
/* GdkFont*				textfont; */
extern char 			*optarg;
extern int 				optind, opterr, optopt;
typeutil					*tl;
int				fullupdate=7000; /* 1/1000 seconds */
int screenWidth, screenHeight;
/* END globals  ----- ----- ----- ----- ----- ----- ----- ----- ----- ----- ----- -----*/


updates_t *new_updates_t()
{
	updates_t	*r;
	r=(updates_t *)malloc(sizeof(updates_t));
	if (r)
	{
		memset(r,0,sizeof(updates_t));
		return r;
	}
	else
	{
		return NULL;
	};
}

clients_t *new_clients_t()
{
	clients_t	*r;
	r=(clients_t *)malloc(sizeof(clients_t));
	if (r)
	{
		memset(r,0,sizeof(clients_t));
		return r;
	}
	else
	{
		return NULL;
	};
}

cameras_t *new_cameras_t()
{
	cameras_t	*r;
	r=(cameras_t *)malloc(sizeof(cameras_t));
	if (r)
	{
		memset(r,0,sizeof(cameras_t));
		return r;
	}
	else
	{
		return NULL;
	};
};

camera_t *new_camera_t()
{
	camera_t	*r;
	r=(camera_t *)malloc(sizeof(camera_t));
	if (r)
	{
		memset(r,0,sizeof(camera_t));
		return r;
	}
	else
	{
		return NULL;
	};
};

client_t *new_client_t()
{
	client_t	*r;
	r=(client_t *)malloc(sizeof(client_t));
	if (r)
	{
		memset(r,0,sizeof(client_t));
		r->sending=FALSE;
		return r;
	}
	else
	{
		return NULL;
	};
};

/* compacts queue, returns new request parameters */
updates_t *optimize_queue(client_t *client)
{
	updates_t	*ret=NULL;
	if 	( client->num_updates > MAX_UPDATE_QUEUE )
	{
		int 	minx=screenWidth,
					miny=screenHeight,
					maxx=0,
					maxy=0,
					cnt=0,
					tx,ty;
		updates_t *list;
		updates_t *i,*j, *u, *v;

		ret=(updates_t *)malloc(sizeof(updates_t));

		/* we're behind, so clear list and replace with one update of the whole area */
		printf("compacting updates\n");
		u=client->updates;
		
		while (u)
		{
			cnt++;
			v=u->next;
			tx=u->x+u->w;
			ty=u->y+u->h;
			if (u->x < minx) {minx = u -> x;};
			if (u->y < miny) {miny = u -> y;};
			if (tx > maxx) {maxx = tx;};
			if (ty > maxy) {maxy = ty;};
			free(u);
			u=v;
		};
		client->updates=list=NULL;
		client->num_updates=0;
		/* new coordinates for larger update */
		ret->x=minx;
		ret->y=miny;
		ret->w=maxx-minx;
		ret->h=maxy-miny;
		printf("** combined %d updates into %dx%d @ %d,%d\n",cnt,ret->w,ret->h,ret->x,ret->y);
	};
	return ret;
}; /* optimize_queue */


/* updates SMALL_TILE sized squared of the screen */
void add_update(client_t *client, int x, int y, int w, int h, int type)
{
	updates_t *list;
	updates_t *i,*j, *u, *v, *oq;
	int flag=0;
	int t1x,t1y,t2x,t2y;

	
	if (client==NULL)
	{
		return;
	};

	if (x>=screenWidth)
	{
		return;
	};
	if (y>=screenHeight)
	{
		return;
	};

	/* figure out origin and destination tiles */
	t1x=x/SMALL_TILE;
	t1y=y/SMALL_TILE;
	t2x=(x+w)/SMALL_TILE;
	t2y=(y+h)/SMALL_TILE;

	/* figure new x,y,w,h based on origin and destination tiles */
	x=t1x*SMALL_TILE;
	y=t1y*SMALL_TILE;
	w=(t2x-t1x)*SMALL_TILE+SMALL_TILE;
	h=(t2y-t1y)*SMALL_TILE+SMALL_TILE;

	if (x+w>=screenWidth)
	{
		x=screenWidth-w;
	};

	if (y+h>=screenHeight)
	{
		y=screenHeight-h;
	};
	
	if (x<0) {x=0;};
	if (y<0) {y=0;};
	if (w>screenWidth) {w=screenWidth;};
	if (h>screenHeight) {h=screenHeight;};
	
#ifdef DEBUG
	printf("adding update (%d): %dx%d @ %d,%d\n",type,w,h,x,y);
#endif
	
	
	/* queue optimizer */
	oq=optimize_queue(client);
	if (oq != NULL)
	{
		w=oq->w;
		h=oq->h;
		x=oq->x;
		y=oq->y;
		free(oq);
	};

	list=client->updates;
	flag=0;
	
	if (list==NULL)
	{
		/* new list */
		list=new_updates_t();
		list->x=x;
		list->y=y;
		list->w=w;
		list->h=h;
		list->t=type;
		client->updates=list;
#ifdef DEBUG
		printf("new list: %dx%d @ %d,%d\n",w,h,x,y);
#endif
	}
	else
	{
		i=list;
		
		while(i)
		{
			j=i;
			i=i->next;
		};
		i=new_updates_t();
		j->next=i;
		i->prev=j;
		i->x=x;
		i->y=y;
		i->w=w;
		i->h=h;
		i->t=type;
#ifdef DEBUG
		printf("add to list: %dx%d @ %d,%d\n",w,h,x,y);
#endif
	};
	client->num_updates++;
	/* signal to send new frame to client */
	if (client->client_intialized!=0)
	{
		g_timeout_add(50,do_update,client);
	};
	return;
};

/* be sure to initialize the list pointer to null! */
cameras_t *add_camera(cameras_t *list, camera_t *camera)
{
	cameras_t *h,*i;
	if (camera==NULL)
	{
		return list;
	};
	if (list==NULL)
	{
		/* new list */
		list=new_cameras_t();
		list->this=camera;
	}
	else
	{
		h=list;
		while(h)
		{
			i=h;
			h=h->next;
		};
		h=new_cameras_t();
		h->this=camera;
		i->next=h;
		h->prev=i;
	};
	return list;
};

/* be sure to initialize the list pointer to null! */
clients_t *add_client(clients_t *list, client_t *client)
{
	clients_t *h,*i;

	if (client==NULL)
	{
		return list;
	};

	if (list==NULL)
	{
		/* new list */
		list=new_clients_t();
		list->this=client;
	}
	else
	{
		h=list;
		while(h)
		{
			i=h;
			h=h->next;
		};
		h=new_clients_t();
		h->this=client;
		i->next=h;
		h->prev=i;
	};
	return list;
};

/* search for clients_t node in clientlist, remove it */
void remove_client(client_t *client)
{
	clients_t *h=NULL,
						*i=NULL,
						*j=NULL;
	updates_t	*u=NULL,	
						*v=NULL;
						
	if (client==NULL)
	{
#ifdef DEBUG
		printf("no client, %d\n",__LINE__);
#endif
		return;
	};

	shutdown(client->socket,2);
	close(client->socket);

	if (clientlist==NULL)
	{
#ifdef DEBUG
		printf("no list, %d\n",__LINE__);
#endif
		return;
	};


	/* search for node in question */
	h=clientlist;
	
	while(h)
	{
		i=h;
		if (i->this==client)
		{
#ifdef DEBUG
			printf("** removing client %p -> %p **\n",i,client);
#endif
			break;
		};
		h=h->next;
	};
	h=NULL;

	/* i is the clients_t node in question */
	if (i==NULL)
	{
#ifdef DEBUG
		printf("no client, %d\n",__LINE__);
#endif
		return;
	};
	if (i->prev != NULL)
	{
#ifdef DEBUG
		printf("isprev (%p), %d\n",i->prev,__LINE__);
#endif
		h=i->prev;
	};
	if (i->next != NULL)
	{
#ifdef DEBUG
		printf("isnext, %d\n",__LINE__);
#endif
		j=i->next;
	};

	i->prev=NULL;
	i->next=NULL;

	if (j != NULL)
	{
#ifdef DEBUG
		printf("isnext, %d\n",__LINE__);
#endif
		/* there is a next */
		if (h != NULL)
		{
#ifdef DEBUG
			printf("isprevt, %d\n",__LINE__);
#endif
			/* there is a previous */
			j->prev=h;
			h->next=j;
		}
		else
		{
#ifdef DEBUG
			printf("noprev, %d\n",__LINE__);
#endif
			j->prev=NULL;
			clientlist=j;
		};
	}
	else
	{
#ifdef DEBUG
		printf("nonext, %d\n",__LINE__);
#endif
		/* there is not a next */
		if (h != NULL)
		{
			/* there is a previous */
#ifdef DEBUG
			printf("isprev, %d\n",__LINE__);
#endif
			h->next=NULL;
		}
		else
		{
#ifdef DEBUG
			printf("noprev, %d\n",__LINE__);
#endif
			/* hmmm, there must not be a list at all. */
			clientlist=NULL;
		};
	};
	
	u=client->updates;
	
	while (u)
	{
		v=u->next;
		free(u);
		u=v;
	};
	memset(client,0,sizeof(client_t));
	free(client);
#ifdef DEBUG
	printf("** removed **\n");
#endif
	return;
};




void init_ghttp(camera_t *camera)
{
 	if (camera == NULL)
	{
		return;
	};
	if (camera->request != NULL)
	{
		printf("ERROR: camera already has request object\n");
		return;
	};
	camera->request = ghttp_request_new();
	ghttp_set_sync(camera->request,1);
	ghttp_set_uri(camera->request, camera->url);
	ghttp_set_header(camera->request, http_hdr_Connection, "close"); 
	ghttp_prepare(camera->request);
	camera->timer_id=g_timeout_add(20,check_req,(gpointer)camera);
};






int ReadExact(int sock, char *buf, int len)
{
	int n; 

	
	while (len > 0) 
	{
		n = recv(sock, buf, len, 0);
		if (n > 0) {
			buf += n;
			len -= n;
		} else {
			return n;
		}
	}
	return 1;
}




void dorfbSetPixelFormat(client_t *client)
{
	int n;
	rfbSetPixelFormatMsg *spf;
	if (client==NULL)
	{
		return;
	};
	spf=&client->rfbformat;

	
	if ((n = ReadExact(client->socket, (char *)spf,
				 sizeof(rfbSetPixelFormatMsg))) <= 0) 
	{
		if (n == 0)
			fprintf(stderr,"rfbProcessClientMessage: client gone\n");
		else
			perror("rfbProcessClientMessage: read");

		remove_client(client);
		return;
	}

	spf->format.trueColour = (spf->format.trueColour ? 1 : 0);
	spf->format.bigEndian = (spf->format.bigEndian ? 1 : 0);
	spf->format.redMax = Swap16IfLE(spf->format.redMax);
	spf->format.greenMax = Swap16IfLE(spf->format.greenMax);
	spf->format.blueMax = Swap16IfLE(spf->format.blueMax);

	printf("!!client->rfbformat.format.bitsPerPixel:%d\n",client->rfbformat.format.bitsPerPixel);



}; /* rfbSetPixelFormat */


/*
* WriteExact writes an exact number of bytes on a TCP socket.  Returns 1 if
* those bytes have been written, or -1 if an error occurred (errno is set to
* ETIMEDOUT if it timed out).
*/
int __WriteExact(client_t *client)
{
	int n;
	int part=SEND_CHUNK;
	int sock;
	char *buf=NULL;

	if (client==NULL)
	{
		return -1;
	};
	
	if (client->sending==TRUE)
	{
#ifdef DEBUG
		printf("already sending\n");
#endif
		return -1;
	};

	client->sending=TRUE;
	
//	printf("sending %d bytes for %p\n",client->upd_buffer_len, client);
	
	sock=client->socket;


	
	while (client->upd_buffer_len > 0) {
		if (part > client->upd_buffer_len) { part=client->upd_buffer_len; };
		buf = &client->upd_buffer[client->upd_buffer_pos];
		n = send(sock, buf, part, 0);

		if (n > 0) {
			client->upd_buffer_pos += n;
			client->upd_buffer_len -= n;
#ifdef DEBUG
			printf("send for %p: %d (%d remains)\n",client,n,client->upd_buffer_len);
#endif
		} 
		else if (n == 0) 
		{
			fprintf(stderr,"WriteExact: write returned 0?\n");
			client->sending=FALSE;
			client->upd_buffer_len=0;
			client->upd_buffer_pos=0;
			remove_client(client);
			return -1;
		} 
		else 
		{
			client->sending=FALSE;
			client->upd_buffer_len=0;
			client->upd_buffer_pos=0;
			return n;
		}
		if (client->upd_buffer_len > 0)
		{
			run_g_q();
		};
	} /* while */
	client->sending=FALSE;
	client->upd_buffer_len=0;
	client->upd_buffer_pos=0;
	return 1;
}


/* rotates 24-bit gdkpixbuf 180 degrees */
char *rotate_180(char *buf, int len)
{
	int 	x,y,z;
	char 	t;
	char 	*out;
	int pixels;
	out=(char *)malloc(128+sizeof(char)*len);
	memset(out,0,128+sizeof(char)*len);
	pixels=len/3;
#ifdef DEBUG
	printf("rotating %d bytes of image data\n",len);
#endif
	for (z=0; z<pixels; z++)
	{
		x=z*3;
		for(y=0;y<3;y++)
		{
			out[x+y]=buf[len-x+y];
		};
	};
#ifdef DEBUG
	printf("done\n");
#endif
	return out;
}

gboolean do_interval_update(gpointer data)
{
	client_t	*client=(client_t	*)data;

	if (data==NULL)
	{
		return TRUE;
	};
	if (! is_client_connected(client))
	{
		printf("client not connected, %d\n",__LINE__);
		return FALSE;
	};
	add_update(client,0,0,screenWidth,screenHeight,MSG_IMAGE);
	return TRUE;
};

gboolean is_client_connected(client_t *client)
{
	clients_t	*u,*v;
	int flag=0;
	int x=0;

	if (clientlist==NULL)
	{
#ifdef DEBUG
		printf("no list, %d\n",__LINE__);
#endif
		return FALSE;
	};

	if (client==NULL)
	{
#ifdef DEBUG
		printf("no client, %d\n",__LINE__);
#endif
		return FALSE;
	};

	u=clientlist;
	
	while(u)
	{
#ifdef DEBUG
		printf("list; #%d: %p -> %p\n",x++,u,u->this);
#endif
		if (u->this == client)
		{
			flag=1;
#ifdef DEBUG
			printf("found, %d\n",__LINE__);
#endif
			break;
		};
		u=u->next;
	};

	if (flag==0)
	{
#ifdef DEBUG
		printf("client %p gone\n", client);
#endif
		return FALSE;
	};

#ifdef DEBUG
	printf("client %p here\n", client);
#endif
	return TRUE;
};




gboolean do_update(gpointer data)
{
	client_t *client=(client_t *)data;
	updates_t *i,*j;
	int x;

	if (data==NULL)
	{
		return FALSE;
	};
	
	if (client->sending==TRUE)
	{
		return FALSE;
	};

	/* make sure client is still in list */
	if (is_client_connected(client) == FALSE)
	{
#ifdef DEBUG
		printf("client not connected.\n");
#endif
		return FALSE;
	};


	i=client->updates;
	if (i==NULL)
	{
		return FALSE;
	};

	/* pull update from head of list */
	client->updates=i->next;
	if (client->updates != NULL)
	{
		client->updates->prev=NULL;
	};
	client->num_updates--;
	if (client->num_updates<0)
	{
		client->num_updates=0;
	};

#ifdef DEBUG
	printf("doing update; type:%d; %d left; %d,%d @ %dx%d\n",i->t,client->num_updates,i->w,i->h,i->x,i->y);
#endif
	if (i->t==MSG_IMAGE)
	{
		dosendchunk(i->x,i->y,i->w,i->h,client);
	};
	if (i->t==MSG_COPY)
	{
		dosendcopyrect(i->x,i->y,i->w,i->h,client);
	};
	free(i);

	return TRUE;
};


/* new camera image */
void update_clients(int x, int y, int w, int h)
{
	clients_t	*i;
	client_t *client;
	i=clientlist;
	
	while(i)
	{
		client=i->this;
		add_update(client,x,y,w,h,MSG_IMAGE);
		i=i->next;
	};
	return;
};

void overlay(GdkPixbuf *s, GdkPixbuf *d, int dx, int dy, camera_t *camera)
{
	int fd;
	char *rbuf,*tbuf;
	int w,h;
	GdkPixbuf	*rp;
	char	caption[1024];
	char	timest[1024];
	const struct tm *tm;
	char	filename[1024];
	time_t t;
	
	if (s==NULL || d==NULL)
	{
		printf("could not overlay!\n");
		return;
	};

	w=gdk_pixbuf_get_width(s);
	h=gdk_pixbuf_get_height(s);

	if (camera->rotate==TRUE)
	{
		rbuf=rotate_180(gdk_pixbuf_get_pixels(s),w*h*3);
	}
	else
	{
		tbuf=gdk_pixbuf_get_pixels(s);
		rbuf=(char *)malloc(sizeof(char)*w*h*3);
		rbuf=memcpy(rbuf,tbuf,w*h*3);
	};

	run_g_q();

	time(&t);
	tm=localtime(&t);
	strftime(timest,1023,"%b %d, %Y %r",tm);
	sprintf(caption,"%s %s",camera->name,timest);
	drawstring(tl, caption, rbuf, w, h, 3, 10, 10);

	if (camera->jpegout)
	{
		printf("writing %s\n",camera->jpegout);
		write_JPEG_file(camera->jpegout,75,rbuf,h,w);
	};
	run_g_q();

	rp=gdk_pixbuf_new_from_data (rbuf,GDK_COLORSPACE_RGB,0,8,w,h,w*3,NULL,NULL);
	gdk_pixbuf_copy_area(rp,0,0,w,h,d,dx,dy);


	free(rbuf);
	gdk_pixbuf_unref(rp);
	update_clients(dx,dy,gdk_pixbuf_get_width(s),gdk_pixbuf_get_height(s));
};


gboolean check_req(gpointer data)
{
	camera_t 	*camera = NULL;
	int				fd,bw,id;
	char			filename[1024];
	ghttp_status	ret;

	camera=(camera_t *)data;

	if (camera==NULL)
	{
#ifdef DEBUG
		printf("NULL CAMERA!\n");
#endif
		return FALSE;
	};

	if (camera->request==NULL)
	{
		printf("NULL REQUEST!\n");
		return FALSE;
	};

	ret=ghttp_process(camera->request);

	if (ret == ghttp_error)
	{
		printf("error retrieving image for %s, requeueing\n",camera->name);
			camera->http_p_count=0;
			ghttp_request_destroy(camera->request);
			camera->request=NULL;
			init_ghttp(camera);
			return FALSE;
	};

	if ( ret != ghttp_done)
	{
		/* (http_timeout/http_check_interval) */
		if (camera->http_p_count++ > 50 )
		{
			camera->http_p_count=0;
			ghttp_request_destroy(camera->request);
			camera->request=NULL;
			init_ghttp(camera);
			return FALSE;
		};
		return TRUE;
	}
	else
	{
#ifdef DEBUG
		printf("retrieval complete for %s; %d bytes; %d polls\n", camera->name,ghttp_get_body_len(camera->request),camera->http_p_count);
#endif
		if (ghttp_get_body_len(camera->request)>0)
		{
			sprintf(filename,"cam-%s-tmp.jpg",camera->name);
#ifdef DEBUG
			printf("storing in %s\n",filename);
#endif
			/* 
			I hate this. But the only alternative is the loader.
			*/
			fd=open(filename,O_WRONLY|O_CREAT,0666);
			if (fd>0)
			{
				errno=0;
				bw=write(fd,ghttp_get_body(camera->request),ghttp_get_body_len(camera->request));
#ifdef DEBUG
				printf("wrote %d bytes to \"%s\" (%d)\n",bw,filename,errno);
#endif
				close(fd);

				camera->pixbuf=gdk_pixbuf_new_from_file(filename);
				if (camera->pixbuf != NULL)
				{
					overlay(camera->pixbuf,main_canvas,camera->x,camera->y,camera);
					gdk_pixbuf_unref(camera->pixbuf);
				}
				else
				{
#ifdef DEBUG
					printf("could not load pixbuf\n");
#endif
				};
			}
			else
			{
				printf("could not open \"%s\" for writing\n",filename);
			};
		};

		/* do it all again! */
		camera->http_p_count=0;
		ghttp_request_destroy(camera->request);
		camera->request=NULL;
		init_ghttp(camera);
		return FALSE;
	};
};

void dorfbSetEncodings(int sock)
{
	/* This reads but ignores the encoding list.
	 We simply check here that the client 
	 can do RRE encoding.*/

	rfbSetEncodingsMsg se;
	CARD32 enc;
	int i;
	int candoRRE = 0;
	int n;

	
	if ((n = ReadExact(sock, (char *)&se, sizeof(rfbSetEncodingsMsg))) <= 0) 
	{
		if (n == 0)
			fprintf(stderr,"rfbProcessClientMessage: client gone\n");
		else
			perror("rfbProcessClientMessage: read");
		return; 
	}

	se.nEncodings = Swap16IfLE(se.nEncodings);
	for (i=0; i<se.nEncodings; i++) 
	{
		
		if ((n = ReadExact(sock, (char *)&enc, sizeof(CARD32))) <= 0) 
		{
			if (n == 0)
				fprintf(stderr, "rfbProcessClientMessage: client gone\n");
			else
				perror("rfbProcessClientMessage: read");
			return;
		}
		if (Swap32IfLE(enc) == rfbEncodingRRE)
		{
			candoRRE = 1;
		};
	}
	if (!candoRRE) 
	{
		fprintf(stderr,"Client can't do RRE\n");
		return;
	}
}; /* dorfbSetEncodings */

void run_g_q()
{
	int count=0;
	
	while(g_main_pending())
	{
		g_main_iteration(FALSE);
/*
		if (count++ > 20)
		{
			printf("too many iterations\n");
			break;
		};
*/
	};
};

void dosendscreenblank(client_t *client)
{
	rfbFramebufferUpdateMsg su;
	rfbFramebufferUpdateRectHeader rect;
	rfbRREHeader hdr;
	CARD32 colour=0;
		
//	return;

	printf("sending screen blank for %dx%d @ %d,%d\n",screenWidth,screenHeight,0,0);
	
	su.type   = rfbFramebufferUpdate;
	su.nRects = Swap16IfLE(1);
	client->upd_buffer=(char *)&su;
	client->upd_buffer_len=sz_rfbFramebufferUpdateMsg;
	WriteExact(client);

	rect.encoding = Swap32IfLE(rfbEncodingRRE);
	rect.r.x = Swap16IfLE(0);
	rect.r.y = Swap16IfLE(0);
	rect.r.w = Swap16IfLE(screenWidth);
	rect.r.h = Swap16IfLE(screenHeight);
	client->upd_buffer=(char *)&rect;
	client->upd_buffer_len=sz_rfbFramebufferUpdateRectHeader;
	WriteExact(client);

	hdr.nSubrects = Swap32IfLE(0);
	client->upd_buffer=(char *)&hdr;
	client->upd_buffer_len=sz_rfbRREHeader;
	WriteExact(client);

	client->upd_buffer=(char *)&colour;
	client->upd_buffer_len = (client->rfbformat.format.bitsPerPixel/8);
	WriteExact(client);

	printf("done sending screen blank for %dx%d @ %d,%d\n",screenWidth,screenHeight,0,0);
};

void dosendcopyrect(int x, int y, int w, int h, client_t *client)
{
	rfbFramebufferUpdateMsg 				su;
	rfbFramebufferUpdateRectHeader 	rect;
	rfbRREHeader 										hdr;
	rfbCopyRect											crect;

	crect.srcX=x;
	crect.srcY=y;
	
	su.type   = rfbFramebufferUpdate;
	su.nRects = Swap16IfLE(1);
	client->upd_buffer=(char *)&su;
	client->upd_buffer_len=sz_rfbFramebufferUpdateMsg;
	WriteExact(client);

	rect.r.x = Swap16IfLE(x);
	rect.r.y = Swap16IfLE(y);
	rect.r.w = Swap16IfLE(w);
	rect.r.h = Swap16IfLE(h);
	rect.encoding = Swap32IfLE(rfbEncodingCopyRect);
	client->upd_buffer=(char *)&rect;
	client->upd_buffer_len=sz_rfbFramebufferUpdateRectHeader;
	WriteExact(client);

	client->upd_buffer=(char *)&crect;
	client->upd_buffer_len=sz_rfbCopyRect;
	WriteExact(client);

	return;
};

void dosendchunk(int x, int y, int w, int h, client_t *client)
{
	rfbFramebufferUpdateMsg 				su;
	rfbFramebufferUpdateRectHeader 	rect;
	rfbRREHeader 										hdr;
	CARD8 													colour;
	int 														n,l,len,fd;
	guchar 													*out,*buf;
	int 														rw,gw,bw;
	GdkPixbuf 											*tile;
	GIOChannel											*gc;
	int 														s_p, d_p;

#ifdef DEBUG
	printf("requested %dx%d at %d,%d (%d,%d)\n",w,h,x,y,screenWidth,screenHeight);
#endif	

	su.type   = rfbFramebufferUpdate;
	su.nRects = Swap16IfLE(1);
	client->upd_buffer=(char *)&su;
	client->upd_buffer_len=sz_rfbFramebufferUpdateMsg;
	WriteExact(client);

	rect.r.x = Swap16IfLE(x);
	rect.r.y = Swap16IfLE(y);
	rect.r.w = Swap16IfLE(w);
	rect.r.h = Swap16IfLE(h);
	rect.encoding = Swap32IfLE(rfbEncodingRaw);
	client->upd_buffer=(char *)&rect;
	client->upd_buffer_len=sz_rfbFramebufferUpdateRectHeader;
	WriteExact(client);

#ifdef DEBUG
	printf("copying (%dx%d) at (%d,%d)\n",w,h,x,y);
#endif
	tile=gdk_pixbuf_new(GDK_COLORSPACE_RGB,FALSE,8,w,h);
	gdk_pixbuf_copy_area(main_canvas,x,y,w,h,tile,0,0);

/*
	buf=gdk_pixbuf_get_pixels(main_canvas);
*/
	buf=gdk_pixbuf_get_pixels(tile);

	bw=(1<<client->rfbformat.format.greenShift)-1;  	/* 2^5-1 */
	gw=(1<<(client->rfbformat.format.redShift))-bw-1;				/* 2^11-2^5 */
	rw=(1<<(client->rfbformat.format.bitsPerPixel))-(1<<client->rfbformat.format.redShift);	/* 2^16-2^11 */
	
		
  client->hformat.indexed=0;
  client->hformat.bits=client->rfbformat.format.bitsPerPixel;
  client->hformat.r=rw; 
  client->hformat.g=gw; 
  client->hformat.b=bw;
  client->hformat.a=0;

#ifdef DEBUG
	printf("1: r: %u  g: %u  b: %u\n",client->rfbformat.format.redShift,
						client->rfbformat.format.greenShift,
						client->rfbformat.format.blueShift);
	printf("2: r: %u  g: %u  b: %u\n",rw,gw,bw);
#endif

	s_p=gdk_pixbuf_get_bits_per_sample(tile)/8*gdk_pixbuf_get_n_channels(tile)*w;
	d_p=(w*client->hformat.bits/8);

	len=h*d_p;
	client->upd_buffer_len=len;
	client->upd_buffer=(unsigned char	*)malloc(client->upd_buffer_len);

	Hermes_ConverterRequest(	conv,
														&fmt_source,
														&client->hformat );

	
#ifdef DEBUG
	printf("w:%d; h:%d; x:%d; y:%d; s:%d; d:%d\n",w,h,x,y,s_p,d_p);
#endif

	Hermes_ConverterCopy(	conv,			/* context */
														buf,			/* pixels in */
														0,				/* origin x*/
														0,				/* origin y */
														w,				/* width */
														h,				/* height */
														s_p,	/*	 source pitch */
														client->upd_buffer, 			/* destination */
														0,				/* origin x*/
														0,				/* origin y */
														w,				/* width */
														h,				/* height */
														d_p	/* bytes per line */ );
												

	
	client->upd_buffer_pos=0;
	WriteExact(client);	
	free(client->upd_buffer);
	gdk_pixbuf_unref(tile);
#ifdef DEBUG
	printf("serviced %dx%d at %d,%d (%d,%d)\n",w,h,x,y,screenWidth,screenHeight);
#endif	
	return;
};

/* client requested a specific update */
void dorfbFramebufferUpdateRequest(client_t *client)
{
	/* This is the important bit.
	 The client has requested an update, so we need to send it
	 the display. */

	rfbFramebufferUpdateRequestMsg sur;
	static counterval = 0;
	static oldcounterval = 0;
	int n;
	int x,y,w,h;

#ifdef DEBUG
	printf("dorfbFramebufferUpdateRequest: %d (%d)\n",client->last_time,client->last_time-time(NULL));
#endif
	
	if ((n = ReadExact(client->socket, (char *)&sur,sizeof(rfbFramebufferUpdateRequestMsg)))<=0) 
	{
		if (n == 0)
		{
			fprintf(stderr,"rfbProcessClientMessage: client gone\n");\
		}
		else
		{
			perror("rfbProcessClientMessage: read error; client gone");
		}
		remove_client(client);
		return;
	}

	if (client->client_intialized==0)
	{
		dosendscreenblank(client);
		client->client_intialized=1;
	};


	x=Swap16IfLE(sur.x);
	y=Swap16IfLE(sur.y);
	w=Swap16IfLE(sur.w);
	h=Swap16IfLE(sur.h);

	if (sur.incremental==0) 
	{
		if (w<screenWidth || h<screenHeight)
		{
#ifdef DEBUG
			printf("**non-incremental: %dx%d at %d,%d \n",w,h,x,y);
#endif
			add_update(client,x,y,w,h,MSG_IMAGE);
		};
		return;
	} 
	else 
	{ 
		if (w<screenWidth || h<screenHeight)
		{
#ifdef DEBUG
			printf("**incremental: %dx%d at %d,%d \n",w,h,x,y);
#endif
			add_update(client,x,y,w,h,MSG_IMAGE);
		};
		return;
	}
}; /* dorfbFramebufferUpdateRequest */



gboolean rfb_cb(GIOChannel *source, GIOCondition condition, gpointer data)
{
	client_t				*client=NULL;
	int 						n,fd;
	CARD8 					msgType;

	fd=g_io_channel_unix_get_fd(source);

	client=(client_t *)data;
	if (client==NULL)
	{
		return FALSE;
	};

//	printf("rfb_cb, %d\n",__LINE__);
	
	if ((n = recv(fd, (char *)&msgType, 1, MSG_PEEK)) <= 0) 
	{
	  	if (n == 0)
			{
				fprintf(stderr,"rfb_cb: client gone\n");
			}
	  	else
			{
				perror("rfb_cb: recv error");
			};
			remove_client(client);
	  	return FALSE;
	}
	
//	printf("message type: %d\n",msgType);
	
	switch (msgType) 
	{
		case rfbSetPixelFormat:
		{
#ifdef DEBUG
			printf("rfbSetPixelFormat\n");
#endif
			dorfbSetPixelFormat(client);
			break;
		};
		case rfbFixColourMapEntries:
		{
			fprintf(stderr,
			"rfbProcessClientMessage: cannot handle colourmaps\n");
			remove_client(client);
			break;
		};
		case rfbSetEncodings:
		{
#ifdef DEBUG
			printf("rfbSetEncodings\n");
#endif
			dorfbSetEncodings(fd);
			break;
		};

		case rfbFramebufferUpdateRequest:
		{
#ifdef DEBUG
			printf("rfbFramebufferUpdateRequest\n");
#endif
			dorfbFramebufferUpdateRequest(client);
			break;
		};

		case rfbKeyEvent:
		{
			rfbKeyEventMsg key;
			
			if ((n = ReadExact(fd, (char *)&key, sizeof(rfbKeyEventMsg))) <= 0) 
			{
				if (n == 0)
					fprintf(stderr,"rfbProcessClientMessage: client gone\n");
				else
					perror("rfbProcessClientMessage: read");
				remove_client(client);
				return;
			}
			key.key = Swap32IfLE(key.key);
			key.key = key.key & 0xff;
#ifdef DEBUG
			printf("key:%c\n",key.key);
#endif
			if (key.key==' ')
			{
				add_update(client,0,0,screenWidth,screenHeight,MSG_IMAGE);
			};
			break;
		};
		case rfbPointerEvent:
		{
			rfbPointerEventMsg ptr;
			
			if ((n = ReadExact(fd, (char *)&ptr, sizeof(rfbPointerEventMsg))) <= 0) 
			{
				if (n == 0)
					fprintf(stderr,"rfbProcessClientMessage: client gone\n");
				else
					perror("rfbProcessClientMessage: read");
				return;
			}
			break;
		};
		case rfbClientCutText:
		{
#ifdef DEBUG
			printf("rfbClientCutText\n");
#endif
			break;
		};
		default:
		{
			fprintf(stderr,"rfbProcessClientMessage: unknown message type %d\n",msgType);
			break;
		};
		
	}; /* switch */	
	return TRUE;
};

gboolean listen_cb(GIOChannel *source, GIOCondition condition, gpointer data)
{
	int 										fd,
													newfd;
	struct sockaddr					addr;
	socklen_t								sl;
	client_t								*client=NULL;
	CARD32 									auth;
	rfbProtocolVersionMsg 	pv;
	GIOChannel 							*gc=NULL;	
	int 										one = 1;
	int											x,y,w,h,numtiles;

/* really no need to sendwhole framebuffer every X seconds */
/*	g_timeout_add(fullupdate,do_interval_update,(gpointer)client); */

#ifdef DEBUG
	printf("---------new incoming connection\n");
#endif
	fd=g_io_channel_unix_get_fd(source);
	newfd=accept(fd,&addr,&sl);
	if (newfd>0)
	{	
#ifdef DEBUG
		printf("accepted\n");
#endif
		setsockopt(newfd, IPPROTO_TCP, TCP_NODELAY, (const char *) &one, sizeof(one));

		client=new_client_t();
		client->socket=newfd;

		clientlist=add_client(clientlist,client);

		/* now, negotiate for pixel format, etc */
		sprintf(pv,rfbProtocolVersionFormat,rfbProtocolMajorVersion, rfbProtocolMinorVersion);
#ifdef DEBUG
	 	printf("Writing protocol version\n");
#endif
		client->upd_buffer=pv;
		client->upd_buffer_len=sz_rfbProtocolVersionMsg;
		if (WriteExact(client) < 0) {
	    	perror("Error writing protocol version");
				remove_client(client);
				return TRUE;
		}

#ifdef DEBUG
   	printf("Reading client protocol version\n");
#endif
		
		if (ReadExact(client->socket, pv,	 sz_rfbProtocolVersionMsg) < 0) {
	    	perror("Error reading client protocol version");
				remove_client(client);
				return TRUE;
		}

		/* No authentication */
		auth = Swap32IfLE(rfbNoAuth);
#ifdef DEBUG
   	printf("Writing NoAuth\n");
#endif
		client->upd_buffer=(char *)&auth;
		client->upd_buffer_len=sizeof(auth);
		if (WriteExact(client) < 0) {
	    	perror("Error writing NoAuth");
				remove_client(client);
				return TRUE;
		}

		/* Read & ignore client Init msg */
#ifdef DEBUG
   	printf("Reading client init msg\n");
#endif
		
		if (ReadExact(client->socket, (char *) &client->cim,	 sz_rfbClientInitMsg) < 0) {
	    	perror("Error reading client init msg");
				remove_client(client);
				return TRUE;
		}

#ifdef DEBUG
	 	printf("Sending server init msg\n");
#endif
		client->upd_buffer=(char *)&sim;
		client->upd_buffer_len=sz_rfbServerInitMsg;
		if (WriteExact(client) < 0) {
	    	perror("Error sending server init msg");
				remove_client(client);
				return TRUE;
		}

#ifdef DEBUG
   	printf("Sending server name\n");
#endif
		client->upd_buffer=servername;
		client->upd_buffer_len=strlen(servername);
		if (WriteExact(client) < 0) {
	    	perror("Error sending server name");
				remove_client(client);
				return TRUE;
		}


		/* listen for messages */
		gc=g_io_channel_unix_new(newfd);
		g_io_add_watch(gc,G_IO_IN,rfb_cb,client);

		numtiles=MAX_UPDATE_QUEUE;
		w=1+numtiles/4;
		h=1+w/2;
		printf("initial update in %d,%d\n",w,h);
		/* send inital update in sections */
		for (y=0; y<=h; y++)
		{
			for (x=0; x<=w; x++)
			{
				add_update(client,x*screenWidth/w,y*screenHeight/h,screenWidth/w,screenHeight/h,MSG_IMAGE);
			};
		};
	};
	return TRUE;
};



#define DO(x) if ((x) < 0) { close(sock); return NULL; }
GIOChannel *ListenAtTcpAddr(unsigned short port)
{
    int sock;
    struct sockaddr_in addr;
    int one = 1;
		GIOChannel *gc=NULL;
	
    addr.sin_family = AF_INET;
    addr.sin_port = htons(port);
    addr.sin_addr.s_addr = INADDR_ANY;
	
    DO(sock = socket(AF_INET, SOCK_STREAM, 0));
    DO(setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, (const char *) &one, sizeof(one)));
    DO(bind(sock, (struct sockaddr *)&addr, sizeof(addr)));
    DO(listen(sock, 5));
		gc=g_io_channel_unix_new(sock);
		g_io_add_watch(gc,G_IO_IN,listen_cb,NULL);
    return gc;
}

/* draws single grayscale-rendered glyphs onto 24-bit RGB bitmap */
void drawletter(typeutil *t, char c, char p, unsigned char *bitmap, int w, int h, int bpp, int ox, int oy)
{
	unsigned char *pBuf;
	int i,j,x,y;
	int pix,color,b,o,px,py;
	int rowstride=w*bpp;
	int szb=h*rowstride;
	int	val=0;
	int ho;
	/* 
		ox and oy are the origin in the bitmap.
		rowstride is the width times the pits-per-pixel of the bitmap.
			I assume that it's a 24-bit RGB packed-pixel format, because I'm
			using GdkPixbuf.
		t->pen_x and t->pen_y are the current drawing offsets for the string.
		This will multiply the RGB values in bitmap by the grayscale values
		in pBuf.
		i and j are the coordinates of pixels in pBuf.
	*/

	px=t->pen_x;
	py=t->pen_y;

	render_char_gray(t,c,p);

	pBuf = t->bitmap->bitmap.buffer;

	for(i = 0; i < t->bitmap->bitmap.rows; i++)
	{
		for(j = 0; j < t->bitmap->bitmap.width; j++)
		{
			val=(int) *pBuf++;
/*
			ho=t->face->bbox.yMax/64 - t->face->bbox.yMin/64 - t->slot->metrics.horiBearingY/64;
*/
			ho=t->slot->metrics.horiBearingY/64 - t->face->bbox.yMin/64;

			/* coordinates in main bitmap */
			x=(px+ox+j) * bpp;
			y=py+oy+i-ho+10;

			/* actual pixel triplet location in linear framebuffer*/
			pix = y*rowstride+x;

			/* adjust each color in this pixel */
			for (color=0; color<3; color++)
			{
				
				b=i*t->bitmap->bitmap.width+j;
				o=pix+color;
				if (pix+color > szb)
				{
					printf("out of bounds at %d,%d\n",x,y);
				}
				else
				{
					bitmap[o]=( bitmap[o] * (255-val) )/255;
				} /* if */
			}; /* for */
		} /* for */
	} /* for */
	FT_Done_Glyph(t->glyph);
}; /* drawletter */

/* draws string of grayscale-rendered glyphs onto 24-bit RGB bitmap */
void drawstring(typeutil *t, char *text, unsigned char *bitmap, int w, int h, int bpp, int x, int y)
{
	int l;
	char c=0;
	char p=0;
	if (t==NULL)
	{
		return;
	};
	/* todo -- blank background area before drawing */
	/* set start position offsets */
	t->pen_x=0;
	t->pen_y=0;
	for (l=0; l<strlen(text); l++)
	{
		c=text[l];
		if (l>0)
		{
			p=text[l-1];
		};
		drawletter(t,c,p,bitmap,w,h,bpp,x,y);
	};
};



/*><*><*><*><*><*><*><*><*><*><*><*><*><*><*><*><*><*><*><*><*><*><*><*><*><*><*><*
	 MAIN
*><*><*><*><*><*><*><*><*><*><*><*><*><*><*><*><*><*><*><*><*><*><*><*><*><*><*><*/

int main(int argc, char **argv)
{
	GMainLoop				*mainloop;
	xmlNodePtr			xp,cxp;
	camera_t 				*camera=NULL;
	int							display; /* VNC display number */
  sigset_t        block_signals;
	char						*configfile=NULL;
	const char 			*optstring="f:";
	char						opt;
	char						*fontname=NULL;
	int							fontsize=12;

	camlist=NULL;
	display=23; /* default */

	setbuf(stdout,NULL);

	while ( (opt=getopt(argc,argv,optstring)) >=0)
	{
		if (opt=='f')
		{
			configfile=strdup(optarg);
		};
	};

	if (configfile==NULL)
	{
		configfile=strdup("vnccam.xml");
	};
	printf("Using %s for configuration\n", configfile);

  /* catch sigpipe */
  sigemptyset(&block_signals);
  sigaddset(&block_signals,SIGPIPE);
  sigprocmask(SIG_SETMASK,&block_signals,NULL);

  if (!Hermes_Init()) 
	{
    printf("Couldn't initialise Hermes!\n");
    exit(1);
  }

	conv=Hermes_ConverterInstance(HERMES_CONVERT_NORMAL);
  if (!conv) {
    printf("Could not obtain converter instance from Hermes!\n");
    exit(1);
  }

	tl=new_typeutil();
	if (tl==NULL)
	{
    exit(1);
	};


	mainloop=g_main_new(FALSE);

	prefs=NULL;
	prefs=xmlParseFile(configfile);
	xp=xmlDocGetRootElement(prefs);
	cxp=xp->xmlChildrenNode;

	while(cxp!=NULL)
	{
		if (!strcmp(cxp->name, "camera"))
		{
			char *tmp;
			camera=new_camera_t();
			camera->url=strdup(xmlGetProp(cxp,"url"));
			camera->name=strdup(xmlGetProp(cxp,"name"));
			camera->x=atol(xmlGetProp(cxp,"x"));
			camera->y=atol(xmlGetProp(cxp,"y"));
			tmp=xmlGetProp(cxp,"jpegout");
			if (tmp)
			{
				camera->jpegout=strdup(tmp);
			};
			if ( xmlGetProp(cxp,"rotate") )
			{
				if ( strcmp(xmlGetProp(cxp,"rotate"),"true") == 0 )
				{
					camera->rotate=TRUE;
					printf("rotating image 180 degrees for camera \"%s\"\n",camera->name);
				}
				else
				{
					camera->rotate=FALSE;
				};
			}
			else
			{
				camera->rotate=FALSE;
			};
			camlist=add_camera(camlist,camera);
			init_ghttp(camera);
/*
			printf("%s: %s\n",cxp->name,camera->name);
			printf("\turl:%s\n",camera->url);
			printf("\tx:%d\n",camera->x);
			printf("\ty:%d\n",camera->y);
*/
		};
		if (!strcmp(cxp->name, "background"))
		{
			printf("background: %s\n",xmlGetProp(cxp,"file"));
			main_canvas=gdk_pixbuf_new_from_file(xmlGetProp(cxp,"file"));
			printf("canvas: w: %d h: %d\n",gdk_pixbuf_get_width(main_canvas),gdk_pixbuf_get_height(main_canvas));
		};
		if (!strcmp(cxp->name, "display"))
		{
			display=atol(xmlGetProp(cxp,"id"));
			servername=strdup(xmlGetProp(cxp,"name"));
			printf("display #%d (port #%d)\n",display,display+RFB_PORT_OFFSET);
		};
		if (!strcmp(cxp->name, "font"))
		{
			printf("setting font\n");
			fontname=strdup(xmlGetProp(cxp,"file"));
			fontsize=atol(xmlGetProp(cxp,"size"));
		};
		if (!strcmp(cxp->name, "fullupdate"))
		{
			fullupdate=atol(xmlGetProp(cxp,"seconds"))*1000;
			printf("Full update every %d seconds\n", fullupdate/1000);
		};
		cxp=cxp->next;
	};

	if (fontname==NULL)
	{
		fontname=strdup("verdana.ttf");
	};
	printf("Using %d point \"%s\" for the font\n",fontsize,fontname);

	set_font(tl,fontname);
	if (tl->face==NULL)
	{
		exit(2);
	};
	set_font_point_size(tl,fontsize);


	screenWidth=gdk_pixbuf_get_width(main_canvas);
	screenHeight=gdk_pixbuf_get_height(main_canvas);

	/* Server Init msg */
	sim.framebufferWidth = Swap16IfLE(screenWidth);
	sim.framebufferHeight = Swap16IfLE(screenHeight);
	sim.format = True16Format;
	sim.format.redMax = Swap16IfLE(sim.format.redMax);
	sim.format.greenMax = Swap16IfLE(sim.format.greenMax);
	sim.format.blueMax = Swap16IfLE(sim.format.blueMax);
	sim.nameLength = strlen(servername);
	sim.nameLength = Swap32IfLE(sim.nameLength);

	/* source format for Hermes */
  fmt_source.indexed=0;
  fmt_source.bits=24;
  fmt_source.r=0x0000ff;
  fmt_source.g=0x00ff00;
  fmt_source.b=0xff0000;
  fmt_source.a=0;

		
	ListenAtTcpAddr(RFB_PORT_OFFSET+display);
	
	g_main_run(mainloop);

	return 0;
};
